/************************************************
*
* JCH2k I2C Library for simultanious master & slave operation
*
* (c) 2006, 2007 J.Claudius Hagen, www.jch2k.de
*
* Revision          : $Revision: 680 $
* Date              : $Date: 2007-03-07 01:51:26 +0100 (Mi, 07 Mrz 2007) $
* Created at        : June 2007
* Updated by        : $Author: claudius $
*
* Details: http://techtalk.jch2k.de (german)
*
**************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>
#include "./twi_masterslave.h"

uint8_t twiBuffer[ TWI_BUFFER_SIZE ];    // Transceiver buffer
uint8_t twiBufferPtr = 0;
uint8_t twiMsgSize = 0;                   // Number of bytes to be transmitted.
uint8_t twiMode = SLAVE;
uint8_t twiIsGencall = 0;
uint8_t twiBusy = 0;





void twiStart(uint8_t msgSize )
{
	twiBusy = BUSY;
  twiMsgSize = msgSize;
  
  TWCR = (1<<TWEN)|
         (1<<TWIE)|(0<<TWINT)|
         (0<<TWEA)|(1<<TWSTA)|(0<<TWSTO)|
         (0<<TWWC);
         
  twiMode = MASTER;
}

void twiInit(uint8_t adress, uint8_t bitrate, uint8_t gencall)
{
  TWBR = bitrate;
  TWSR = 0; // prescaler
  TWAR = (adress<<1);
  
  if (gencall)
  	TWAR |= 1;
  	
  TWDR = 0xFF;                                      // Default content = SDA released.
  
  TWCR = (1<<TWEN)|                             // TWI Interface enabled.
         (1<<TWIE)|(0<<TWINT)|                  // Enable TWI Interupt and clear the flag.
         (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|       // Prepare to ACK next time the Slave is addressed.
         (0<<TWWC);                              //
  
  twiMode = SLAVE;
}    
    


// ********** Interrupt Handlers ********** //

ISR(TWI_vect)
{
  switch (TWSR & 0xF8)
  {
  	//=========
  	// SLAVE CASES
  	
    case TWI_STX_ADR_ACK:                   // Own SLA+R has been received; ACK has been returned
//    case TWI_STX_ADR_ACK_M_ARB_LOST:      // Arbitration lost in SLA+R/W as Master; own SLA+R has been received; ACK has been returned
			twiBusy = BUSY;
			stxStarted();
      twiBufferPtr   = 0;
    case TWI_STX_DATA_ACK:                 // Data byte in TWDR has been transmitted; ACK has been received
    	TWDR = twiBuffer[twiBufferPtr++];
      TWCR = (1<<TWEN)|
             (1<<TWIE)|(1<<TWINT)|
             (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)| 
             (0<<TWWC);
      break;
    case TWI_STX_DATA_NACK:          // Data byte in TWDR has been transmitted; NACK has been received.
      TWCR = (1<<TWEN)|
             (1<<TWIE)|(1<<TWINT)|
             (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|
             (0<<TWWC);
      twiBusy = 0;
      stxComplete();
      break;     
    case TWI_SRX_GEN_ACK:            // General call address has been received; ACK has been returned
//    case TWI_SRX_GEN_ACK_M_ARB_LOST: // Arbitration lost in SLA+R/W as Master; General call address has been received; ACK has been returned
      twiIsGencall = 0xFF;
    case TWI_SRX_ADR_ACK:            // Own SLA+W has been received ACK has been returned
//    case TWI_SRX_ADR_ACK_M_ARB_LOST: // Arbitration lost in SLA+R/W as Master; own SLA+W has been received; ACK has been returned
			twiBusy = BUSY;
      twiBufferPtr   = 0;
      
      TWCR = (1<<TWEN)|
             (1<<TWIE)|(1<<TWINT)|
             (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|
             (0<<TWWC);      
      break;
    case TWI_SRX_ADR_DATA_ACK:       // Previously addressed with own SLA+W; data has been received; ACK has been returned
    case TWI_SRX_GEN_DATA_ACK:       // Previously addressed with general call; data has been received; ACK has been returned
      
      twiBuffer[twiBufferPtr++]     = TWDR;
      
      if (twiIsGencall)
      	gcxByteReceived(TWDR);
      else
      	srxByteReceived(TWDR);
      
      TWCR = (1<<TWEN)|
             (1<<TWIE)|(1<<TWINT)|
             (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|
             (0<<TWWC);
      break;
    case TWI_SRX_STOP_RESTART:       // A STOP condition or repeated START condition has been received while still addressed as Slave    
      
      TWCR = (1<<TWEN)|
             (1<<TWIE)|(1<<TWINT)|
             (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|
             (0<<TWWC);
      twiIsGencall = 0;
      twiBusy = 0;
      srxComplete(twiBufferPtr);
      break;
    
    //==========
    // MASTER CASES           

    case TWI_START:                     // START has been transmitted  
    case TWI_REP_START:                 // Repeated START has been transmitted
      twiBufferPtr = 0;
    case TWI_MTX_ADR_ACK:               // SLA+W has been tramsmitted and ACK received
    case TWI_MTX_DATA_ACK:              // Data byte has been tramsmitted and ACK received
    {
      if (twiBufferPtr < twiMsgSize)
      {
        TWDR = twiBuffer[twiBufferPtr++];
        TWCR = (1<<TWEN)|
               (1<<TWIE)|(1<<TWINT)|
               (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|
               (0<<TWWC);
      }else                    // Send STOP after last byte
      {
        TWDR = 0; 
        TWCR = (1<<TWEN)|
               (1<<TWIE)|(1<<TWINT)|
               (1<<TWEA)|(0<<TWSTA)|(1<<TWSTO)|
               (0<<TWWC);
        twiMode = SLAVE;
        twiBusy = 0;
        mtxComplete();
      }
      break;
  	}
  	
    case TWI_MRX_DATA_ACK:      // Data byte has been received and ACK tramsmitted
      twiBuffer[twiBufferPtr++] = TWDR;
      mrxByteReceived(TWDR);
    case TWI_MRX_ADR_ACK:       // SLA+R has been tramsmitted and ACK received
      if (twiBufferPtr < (twiMsgSize-1) )                  // Detect the last byte to NACK it.
      {
        TWCR = (1<<TWEN)|
               (1<<TWIE)|(1<<TWINT)|
               (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|
               (0<<TWWC);  
      }else                    // Send NACK after next reception
      {
        TWCR = (1<<TWEN)|
               (1<<TWIE)|(1<<TWINT)|
               (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|
               (0<<TWWC);
      }    
      break; 
      
    case TWI_MRX_DATA_NACK:     // Data byte has been received and NACK tramsmitted
      twiBuffer[twiBufferPtr] = TWDR;
      TWCR = (1<<TWEN)|
             (1<<TWIE)|(1<<TWINT)|
             (1<<TWEA)|(0<<TWSTA)|(1<<TWSTO)|
             (0<<TWWC);
      
      twiBusy = 0;
      twiMode = SLAVE;
      mrxComplete(twiBufferPtr);
      break;      
      
    case TWI_ARB_LOST:          // Arbitration lost --> restart
      TWCR = (1<<TWEN)|
             (1<<TWIE)|(1<<TWINT)|
             (0<<TWEA)|(1<<TWSTA)|(0<<TWSTO)|
             (0<<TWWC);
      break;
      
    //======
    // ERROR CASES
      
    case TWI_SRX_ADR_DATA_NACK:      // Previously addressed with own SLA+W; data has been received; NOT ACK has been returned
    case TWI_SRX_GEN_DATA_NACK:      // Previously addressed with general call; data has been received; NOT ACK has been returned
    case TWI_STX_DATA_ACK_LAST_BYTE: // Last data byte in TWDR has been transmitted (TWEA = 0); ACK has been received
		case TWI_MTX_ADR_NACK:      // SLA+W has been tramsmitted and NACK received
    case TWI_MRX_ADR_NACK:      // SLA+R has been tramsmitted and NACK received    
    case TWI_MTX_DATA_NACK:     // Data byte has been tramsmitted and NACK received
//    case TWI_NO_STATE              // No relevant state information available; TWINT = 0
    case TWI_BUS_ERROR:         // Bus error due to an illegal START or STOP condition
    default:     
      TWCR = (1<<TWEN)|
             (1<<TWIE)|(1<<TWINT)|
             (1<<TWEA)|(0<<TWSTA)|(1<<TWSTO)|
             (0<<TWWC);
      twiIsGencall = 0;
      twiMode = SLAVE;
      twiBusy = 0;
      twiError(TWSR & 0xF8);
  }
}
